import math
import random
from typing import Literal, Annotated, override

from pydantic import BaseModel, Field, field_validator, model_validator

from .consts import MODELS, ACTIONS, SAMPLERS, NOISES, RESOLUTIONS


class Metadata(BaseModel):
    """
    Serve as `input` and `parameters` in the request body of the /ai/generate-image endpoint.

    Except for the provided fields, other arbitrary fields can also be added into parameters,
    which will be present in the metadata of the image, like qualityToggle or ucPreset.
    These fields are mainly for the website to remember some key information when importing the image.

    Parameters
    ----------
    | General
    prompt: `str`
        Text prompt to generate image from. Serve as `input` in the request body.
        Refer to https://docs.novelai.net/image/tags.html, https://docs.novelai.net/image/strengthening-weakening.html
    model: `str`, optional
        Model to use for the generation. Serve as `model` in the request body. Refer to `novelai.consts.MODELS`
    action: `str`, optional
        Action to perform. Serve as `action` in the request body. Refer to `novelai.consts.ACTIONS`

    | Note: all fields below that are not in "General" section will together serve as `parameters` in the request body

    | Prompt
    negative_prompt: `str`, optional
        Refer to https://docs.novelai.net/image/undesiredcontent.html
    qualityToggle: `bool`, optional
        Whether to automatically append quality tags to the prompt. Refer to https://docs.novelai.net/image/qualitytags.html
    ucPreset: `int`, optional
        Preset value of undisired content. Refer to https://docs.novelai.net/image/undesiredcontent.html
        Range: 0-3, 0: Heavy, 1: Light, 2: Human Focus, 3: None

    | Image settings
    width: `int`, optional
        Width of the image to generate in pixels
    height: `int`, optional
        Height of the image to generate in pixels
    n_samples: `int`, optional
        Number of images to return. The maximum value is 8 up to 360,448 px, 6 up to 409,600 px,
        4 up to 1,310,720 px, 2 up to 1,572,864 px, and 1 up to the max size

    | AI settings
    steps: `int`, optional
        Refer to https://docs.novelai.net/image/stepsguidance.html
    scale: `float`, optional
        Value of prompt guidance. Refer to https://docs.novelai.net/image/stepsguidance.html
    dynamic_thresholding: `bool`, optional
        Whether to enable descrisper. Refer to https://docs.novelai.net/image/stepsguidance.html#decrisper
    seed: `int`, optional
        Random seed to use for the image (between 0 and 4294967295), defaults to 0
        Seed 0 means that a random seed will be chosen, but not set in the metadata (so giving a seed yourself is important)
        Note: When generating multiple images, each consecutive image adds 1 to the seed parameter.
        This means it can go beyond the limit of 4294967295, making it unreproducible with a single generation
    extra_noise_seed: `int`, optional
        Unknown, but seems to be fulfill the same role as seed
    sampler: `str`, optional
        Refer to https://docs.novelai.net/image/sampling.html.
        The full list of samplers is (some may not work due to being deprecated): "k_lms", "k_euler",
        "k_euler_ancestral", "k_heun", "plms", "ddim", "nai_smea", "nai_smea_dyn", "k_dpmpp_2m",
        "k_dpmpp_2s_ancestral", "k_dpmpp_sde", "k_dpm_2", "k_dpm_2_ancestral", "k_dpm_adaptive", "k_dpm_fast"
    sm: `bool`, optional
        Whether to enable SMEA. Refer to https://docs.novelai.net/image/sampling.html#special-samplers-smea--smea-dyn
    sm_dyn: `bool`, optional
        Whether to enable SMEA DYN (SMEA needs to be enabled for SMEA DYN)
        Refer to https://docs.novelai.net/image/sampling.html#special-samplers-smea--smea-dyn
    uncond_scale: `float`, optional
        Range: 0-1.5, Undesired content strength, refer to https://docs.novelai.net/image/undesiredcontent.html#undesired-content-strength
    cfg_rescale: `float`, optional
        Range: 0-1, Prompt guidance rescale, refer to https://docs.novelai.net/image/stepsguidance.html#prompt-guidance-rescale
    noise_schedule: `str`, optional
        Noise schedule, choose from "native", "karras", "exponential", and "polyexponential"

    | img2img
    image: `str`, optional
        Base64-encoded PNG image for Image to Image
    strength: `float`, optional
        Range: 0.01-0.99, refer to https://docs.novelai.net/image/strengthnoise.html
    noise: `float`, optional
        Range: 0-0.99, refer to https://docs.novelai.net/image/strengthnoise.html
    controlnet_strength: `float`, optional
        Range: 0.1-2, controls how much influence the ControlNet has on the image
    controlnet_condition: `str`, optional
        Base64-encoded PNG ControlNet mask retrieved from the /ai/annotate-image endpoint
    controlnet_model: `str`, optional
        Control tool to use for the ControlNet. Note: V3 model does not support control tools
        Palette Swap is "hed", Form Lock is "depth", Scribbler is "scribble", Building Control is "mlsd", is Landscaper is "seg"

    | Inpaint
    add_original_image: `bool`, optional
        Refer to https://docs.novelai.net/image/inpaint.html#overlay-original-image
    mask: `str`, optional
        Base64-encoded black and white PNG image to use as a mask for inpainting
        White is the area to inpaint and black is the rest

    | Misc
    params_version: `int`, optional
        Defaults to 1
    legacy: `bool`, optional
        Defaults to False
    legacy_v3_extend: `bool`, optional
        Defaults to False
    """

    # General
    # Fields in this section will be excluded from the output of model_dump during serialization
    prompt: str = Field(exclude=True)
    model: str = Field(default=MODELS.V3, exclude=True)
    action: str = Field(default=ACTIONS.GENERATE, exclude=True)

    # Prompt
    negative_prompt: str = ""
    qualityToggle: bool = True
    ucPreset: Literal[0, 1, 2, 3] = 2

    # Image settings
    width: Annotated[int, Field(ge=64, le=49152)] = 1024
    height: Annotated[int, Field(ge=64, le=49152)] = 1024
    n_samples: Annotated[int, Field(ge=1, le=8)] = 1

    # AI settings
    steps: int = Field(default=28, ge=1, le=50)
    scale: float = Field(default=6.0, ge=0, le=10, multiple_of=0.1)
    dynamic_thresholding: bool = False
    seed: int = Field(
        default_factory=lambda: random.randint(0, 4294967295 - 7),
        gt=0,
        le=4294967295 - 7,
    )
    extra_noise_seed: int = Field(
        default_factory=lambda: random.randint(0, 4294967295 - 7),
        gt=0,
        le=4294967295 - 7,
    )
    sampler: str = SAMPLERS.EULER
    sm: bool = True
    sm_dyn: bool = False
    uncond_scale: float = Field(default=1.0, ge=0, le=1.5, multiple_of=0.05)
    cfg_rescale: float = Field(default=0, ge=0, le=1, multiple_of=0.02)
    noise_schedule: str = NOISES.NATIVE

    # img2img
    image: str | None = None
    strength: float | None = Field(default=0.3, ge=0.01, le=0.99, multiple_of=0.01)
    noise: float | None = Field(default=0, ge=0, le=0.99, multiple_of=0.01)
    controlnet_strength: float = Field(default=1, ge=0.1, le=2, multiple_of=0.1)
    controlnet_condition: str | None = None
    controlnet_model: str | None = None

    # Inpaint
    add_original_image: bool = True
    mask: str | None = None

    # Misc
    params_version: Literal[1] = 1
    legacy: bool = False
    legacy_v3_extend: bool = False

    @field_validator("model")
    @classmethod
    def __model_validator(cls, v: str) -> str:
        if v not in MODELS.values():
            raise ValueError(f"Model should be in {list(MODELS.values())}")
        return v

    @field_validator("action")
    @classmethod
    def __action_validator(cls, v: str) -> str:
        if v not in ACTIONS.values():
            raise ValueError(f"Action should be in {list(ACTIONS.values())}")
        return v

    @field_validator("sampler")
    @classmethod
    def __sampler_validator(cls, v: str) -> str:
        if v not in SAMPLERS.values():
            raise ValueError(f"Sampler should be in {list(SAMPLERS.values())}")
        return v

    @field_validator("noise_schedule")
    @classmethod
    def __noise_schedule_validator(cls, v: str) -> str:
        if v not in NOISES.values():
            raise ValueError(f"Noise schedule should be in {list(NOISES.values())}")
        return v

    @model_validator(mode="after")
    def __n_samples_validator(self) -> "Metadata":
        max_n_samples = self.get_max_n_samples()
        if self.n_samples > max_n_samples:
            raise ValueError(
                f"Max value of n_samples is {max_n_samples} under current resolution ({self.width}x{self.height}). Got {self.n_samples}."
            )
        return self

    @override
    def model_post_init(self, *args) -> None:
        """
        Post-initialization hook. Inherit from `pydantic.BaseModel`.
        Implement this method to add custom initialization logic.
        """

        match self.ucPreset:
            case 0:  # Heavy
                self.negative_prompt += ", lowres, {bad}, error, fewer, extra, missing, worst quality, jpeg artifacts, bad quality, watermark, unfinished, displeasing, chromatic aberration, signature, extra digits, artistic error, username, scan, [abstract]"
            case 1:  # Light
                self.negative_prompt += ", lowres, jpeg artifacts, worst quality, watermark, blurry, very displeasing"
            case 2:  # Human Focus
                self.negative_prompt += ", lowres, {bad}, error, fewer, extra, missing, worst quality, jpeg artifacts, bad quality, watermark, unfinished, displeasing, chromatic aberration, signature, extra digits, artistic error, username, scan, [abstract], bad anatomy, bad hands, @_@, mismatched pupils, heart-shaped pupils, glowing eyes"

        if self.qualityToggle:
            self.prompt += ", best quality, amazing quality, very aesthetic, absurdres"

    def get_max_n_samples(self) -> int:
        """
        Get the max allowed number of images to generate in a single request by resolution.

        Returns
        -------
        `int`
            Maximum value of `ImageParams.n_samples`
        """

        w, h = self.width, self.height

        if w * h <= 512 * 704:
            return 8

        if w * h <= 640 * 640:
            return 6

        if w * h <= 1024 * 3072:
            return 4

        return 0

    def calculate_cost(self, is_opus: bool = False):
        """
        Calculate the Anlas cost of current parameters.

        Parameters
        ----------
        is_opus: `bool`, optional
            If the subscription tier is Opus. Opus accounts have access to free generations.
        """

        steps: int = self.steps
        n_samples: int = self.n_samples
        uncond_scale: float = self.uncond_scale
        strength: float = self.action == ACTIONS.IMG2IMG and self.strength or 1.0
        smea_factor = self.sm_dyn and 1.4 or self.sm and 1.2 or 1.0
        resolution = max(self.width * self.height, 65536)

        # For normal resolutions, squre is adjusted to the same price as portrait/landscape
        if resolution > math.prod(
            RESOLUTIONS.NORMAL_PORTRAIT
        ) and resolution <= math.prod(RESOLUTIONS.NORMAL_SQUARE):
            resolution = math.prod(RESOLUTIONS.NORMAL_PORTRAIT)

        per_sample = (
            math.ceil(
                2951823174884865e-21 * resolution
                + 5.753298233447344e-7 * resolution * steps
            )
            * smea_factor
        )
        per_sample = max(math.ceil(per_sample * strength), 2)

        if uncond_scale != 1.0:
            per_sample = math.ceil(per_sample * 1.3)

        opus_discount = (
            is_opus
            and steps <= 28
            and (resolution <= math.prod(RESOLUTIONS.NORMAL_SQUARE))
        )

        return per_sample * (n_samples - int(opus_discount))



import os
import asyncio
from pathlib import Path

from loguru import logger

from novelai import NAIClient, Metadata, HOSTS


client = NAIClient(
    username=os.getenv("USERNAME"), password=os.getenv("PASSWORD")
)

path = Path("./temp")
path.mkdir(parents=True, exist_ok=True)


@logger.catch
async def task_api():
    metadata = Metadata(
        prompt=os.getenv("PROMPT"),
        negative_prompt=os.getenv("NEGATIVE_PROMPT"),
    )

    output = await client.generate_image(
        metadata, host=HOSTS.API, verbose=True, is_opus=True
    )

    for filename, data in output.items():
        dest = Path(path / f"api_{filename}")
        dest.write_bytes(data)
        logger.info(f"Image saved as {dest.resolve()}")

    logger.success("API task completed!")


@logger.catch
async def task_web():
    metadata = Metadata(
        prompt=os.getenv("PROMPT"),
        negative_prompt=os.getenv("NEGATIVE_PROMPT"),
        qualityToggle=False,
        ucPreset=3,
        width=832,
        height=1216,
        n_samples=1,
        sm_dyn=True,
    )

    output = await client.generate_image(
        metadata, host=HOSTS.WEB, verbose=True, is_opus=True
    )

    for filename, data in output.items():
        dest = Path(path / f"web_{filename}")
        dest.write_bytes(data)
        logger.info(f"Image saved as {dest.resolve()}")

    logger.success("Web task completed!")


@logger.catch
async def main():
    await client.init(timeout=45)
    tasks = [
        asyncio.create_task(task_api()),
        asyncio.create_task(task_web()),
    ]
    await asyncio.wait(tasks)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import unittest
from unittest.mock import MagicMock, patch

from novelai import NAIClient, APIError, AuthError, NovelAIError, HOSTS, ENDPOINTS


class TestInit(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.naiclient = NAIClient("test_username", "test_password")

    @patch("novelai.client.AsyncClient.post")
    @patch("novelai.client.encode_access_key")
    async def test_success(self, mock_encode_access_key, mock_post):
        # Mocks
        mock_encode_access_key.return_value = "test_key"
        mock_post.return_value.status_code = 201
        mock_post.return_value.json = MagicMock(
            return_value={"accessToken": "test_token"}
        )

        # Function call
        await self.naiclient.init(timeout=33, auto_close=True, close_delay=0)

        # Assertions
        self.assertEqual(self.naiclient.client.timeout.connect, 33)
        self.assertEqual(self.naiclient.auto_close, True)
        self.assertEqual(self.naiclient.close_delay, 0)
        self.assertEqual(
            self.naiclient.client.headers["Authorization"], "Bearer test_token"
        )
        self.assertTrue(self.naiclient.running)
        mock_post.assert_called_once_with(
            url=f"{HOSTS.API.url}{ENDPOINTS.LOGIN}",
            json={"key": "test_key"},
        )

        # Awaited tasks assertions
        await asyncio.sleep(0.01)
        self.assertFalse(self.naiclient.running)

    @patch("novelai.client.AsyncClient.post")
    async def test_exceptions(self, mock_post):
        # Error cases
        error_codes = [400, 401, 500]
        exceptions = [APIError, AuthError, NovelAIError]

        for code, exception in zip(error_codes, exceptions):
            with self.subTest(f"Status code: {code}"):
                mock_post.return_value.status_code = code

                # Assert exception is raised
                with self.assertRaises(exception):
                    await self.naiclient.init(auto_close=False)

                # Assert AsyncClient is created but closed
                self.assertTrue(self.naiclient.client.is_closed)
                self.assertFalse(self.naiclient.running)


if __name__ == "__main__":
    unittest.main()



import asyncio
from asyncio import Task

from httpx import AsyncClient, ReadTimeout
from pydantic import validate_call
from loguru import logger

from .metadata import Metadata
from .utils import encode_access_key, parse_zip
from .consts import HOSTS, ENDPOINTS, HEADERS
from .types import (
    Host,
    User,
    AuthError,
    APIError,
    NovelAIError,
    TimeoutError,
    ConcurrentError,
)


def running(func) -> callable:
    """
    Decorator to check if client is running before making a request.
    """

    async def wrapper(self: "NAIClient", *args, **kwargs):
        if not self.running:
            await self.init(auto_close=self.auto_close, close_delay=self.close_delay)
            if self.running:
                return await func(self, *args, **kwargs)

            raise Exception(
                f"Invalid function call: NAIClient.{func.__name__}. Client initialization failed."
            )
        else:
            return await func(self, *args, **kwargs)

    return wrapper


class NAIClient:
    """
    Async httpx client interface to interact with NovelAI's service.

    Parameters
    ----------
    username: `str`
        NovelAI username, usually an email address
    password: `str`
        NovelAI password
    proxy: `dict`, optional
        Proxy to use for the client
    """

    __slots__ = [
        "user",
        "proxy",
        "client",
        "running",
        "auto_close",
        "close_delay",
        "close_task",
    ]

    def __init__(
        self,
        username: str,
        password: str,
        proxy: dict | None = None,
    ):
        self.user = User(username=username, password=password)
        self.proxy = proxy
        self.client: AsyncClient | None = None
        self.running: bool = False
        self.auto_close: bool = False
        self.close_delay: int = 0
        self.close_task: Task | None = None

    async def init(
        self, timeout: float = 30, auto_close: bool = False, close_delay: int = 300
    ) -> None:
        """
        Get access token and implement Authorization header.

        Parameters
        ----------
        timeout: `int`, optional
            Request timeout of the client in seconds. Used to limit the max waiting time when sending a request
        auto_close: `bool`, optional
            If `True`, the client will close connections and clear resource usage after a certain period
            of inactivity. Useful for keep-alive services
        close_delay: `int`, optional
            Time to wait before auto-closing the client in seconds. Effective only if `auto_close` is `True`
        """
        try:
            self.client = AsyncClient(
                timeout=timeout, proxies=self.proxy, headers=HEADERS
            )
            self.client.headers["Authorization"] = (
                f"Bearer {await self.get_access_token()}"
            )

            self.running = True
            logger.success("NovelAI client initialized successfully.")

            self.auto_close = auto_close
            self.close_delay = close_delay
            if self.auto_close:
                await self.reset_close_task()
        except Exception:
            await self.close(0)
            raise

    async def close(self, wait: int | None = None) -> None:
        """
        Close the client after a certain period of inactivity, or call manually to close immediately.

        Parameters
        ----------
        wait: `int`, optional
            Time to wait before closing the client in seconds
        """
        await asyncio.sleep(wait is not None and wait or self.close_delay)
        await self.client.aclose()
        self.running = False

    async def reset_close_task(self) -> None:
        """
        Reset the timer for closing the client when a new request is made.
        """
        if self.close_task:
            self.close_task.cancel()
            self.close_task = None
        self.close_task = asyncio.create_task(self.close())

    async def get_access_token(self) -> str:
        """
        Send post request to /user/login endpoint to get user's access token.

        Returns
        -------
        `str`
            NovelAI access token which is used in the Authorization header with the Bearer scheme
        """
        access_key = encode_access_key(self.user)

        response = await self.client.post(
            url=f"{HOSTS.API.url}{ENDPOINTS.LOGIN}",
            json={
                "key": access_key,
            },
        )

        match response.status_code:
            case 201:
                return response.json()["accessToken"]
            case 400:
                raise APIError("A validation error occured.")
            case 401:
                raise AuthError("Invalid username or password.")
            case _:
                raise NovelAIError(
                    f"An unknown error occured. Error message: {response.status_code} {response.reason_phrase}"
                )

    @running
    @validate_call
    async def generate_image(
        self,
        metadata: Metadata | None = None,
        host: Host = HOSTS.API,
        verbose: bool = False,
        is_opus: bool = False,
        **kwargs,
    ) -> dict[str, bytes]:
        """
        Send post request to /ai/generate-image endpoint for image generation.

        Parameters
        ----------
        metadata: `Metadata`
            Metadata object containing parameters required for image generation
        host: `Host`, optional
            Host to send the request. Refer to `novelai.consts.HOSTS` for available hosts or provide a custom host
        verbose: `bool`, optional
            If `True`, will log the estimated Anlas cost before sending the request
        is_opus: `bool`, optional
            Use with `verbose` to calculate the cost based on your subscription tier
        **kwargs: `Any`
            If `metadata` is not provided, these parameters are used to create a `Metadata` object

        Returns
        -------
        `dict`
            Dictionary with file names (`str`) as keys and file contents (`bytes`) as values
        """
        if metadata is None:
            metadata = Metadata(**kwargs)

        if verbose:
            logger.info(
                f"Generating image... estimated Anlas cost: {metadata.calculate_cost(is_opus)}"
            )

        if self.auto_close:
            await self.reset_close_task()

        try:
            response = await self.client.post(
                url=f"{host.url}{ENDPOINTS.IMAGE}",
                json={
                    "input": metadata.prompt,
                    "model": metadata.model,
                    "action": metadata.action,
                    "parameters": metadata.model_dump(exclude_none=True),
                },
            )
        except ReadTimeout:
            raise TimeoutError(
                "Request timed out, please try again. If the problem persists, consider setting a higher `timeout` value when initiating NAIClient."
            )

        match response.status_code:
            case 200:
                assert (
                    response.headers["Content-Type"] == host.accept
                ), f"Invalid response content type. Expected '{host.accept}', got '{response.headers['Content-Type']}'."
                return parse_zip(response.content)
            case 400:
                raise APIError(
                    f"A validation error occured. Message from NovelAI: {response.json().get('message')}"
                )
            case 401:
                self.running = False
                raise AuthError(
                    f"Access token is incorrect. Message from NovelAI: {response.json().get('message')}"
                )
            case 402:
                self.running = False
                raise AuthError(
                    f"An active subscription is required to access this endpoint. Message from NovelAI: {response.json().get('message')}"
                )
            case 409:
                raise NovelAIError(
                    f"A conflict error occured. Message from NovelAI: {response.json().get('message')}"
                )
            case 429:
                raise ConcurrentError(
                    f"A concurrent error occured. Message from NovelAI: {response.json().get('message')}"
                )
            case _:
                raise NovelAIError(
                    f"An unknown error occured. Error message: {response.status_code} {response.reason_phrase}"
                )



import asyncio
from argparse import ArgumentParser

from .client import NAIClient


async def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest='command')

    # Login command
    login_parser = subparsers.add_parser("login", help="Login to NovelAI to get your access token")
    login_parser.add_argument("username", help="NovelAI username")
    login_parser.add_argument("password", help="NovelAI password")

    args = parser.parse_args()

    if args.command == 'login':
        client = NAIClient(username=args.username, password=args.password)
        access_token = await client.get_access_token()
        print(access_token)
    else:
        parser.print_help()


if __name__ == "__main__":
    asyncio.run(main())



import os
import unittest
from unittest.mock import AsyncMock, MagicMock, patch

from httpx import AsyncClient

from novelai import (
    NAIClient,
    APIError,
    AuthError,
    NovelAIError,
    HOSTS,
    ENDPOINTS,
    HEADERS,
)


class TestGetAccessToken(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.naiclient = NAIClient(
            os.getenv("USERNAME") or "test_username",
            os.getenv("PASSWORD") or "test_password",
        )
        self.naiclient.client = AsyncClient(headers=HEADERS)

    @unittest.skipIf(
        not (os.getenv("USERNAME") and os.getenv("PASSWORD")),
        "Skipping test_success...",
    )
    async def test_success(self):
        # Function call
        output_token = await self.naiclient.get_access_token()

        # Send testing request to NovelAI's API
        self.naiclient.client.headers["Authorization"] = f"Bearer {output_token}"
        response = await self.naiclient.client.get(
            url=f"{HOSTS.API.url}{ENDPOINTS.USERDATA}"
        )

        # Assertions
        self.assertEqual(response.status_code, 200)

    @patch("novelai.client.NAIClient.user")
    async def test_auth_error(self, mock_user):
        # Mock the user object to return wrong username and password
        mock_user.return_value = MagicMock(
            username="wrong_username", password="wrong_password"
        )

        # Function call & assertion
        with self.assertRaises(AuthError) as context:
            await self.naiclient.get_access_token()
        self.assertEqual(str(context.exception), "Invalid username or password.")

    @patch("novelai.client.encode_access_key")
    async def test_validation_error(self, mock_encode_access_key):
        # Mock the encode_access_key function to return a wrong string
        mock_encode_access_key.return_value = "misformatted_string"

        # Function call & assertion
        with self.assertRaises(APIError) as context:
            await self.naiclient.get_access_token()
        self.assertEqual(str(context.exception), "A validation error occured.")

    async def test_exceptions(self):
        # Error cases
        error_codes = [400, 401, 500]
        error_exceptions = [APIError, AuthError, NovelAIError]

        for code, exception in zip(error_codes, error_exceptions):
            with self.subTest(f"Status code: {code}"):
                # Mock the AsyncClient's post method to return an error code
                self.naiclient.client.post = AsyncMock(
                    return_value=MagicMock(status_code=code)
                )

                # Check if correct exception is raised
                with self.assertRaises(exception):
                    await self.naiclient.get_access_token()


if __name__ == "__main__":
    unittest.main()



from pydantic import BaseModel
from pydantic.dataclasses import dataclass


class DotDict(dict):
    """
    Dictionary with dot notation access to its keys.
    """
    def __getattr__(self, name):
        return self.get(name, None)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        del self[name]


class User(BaseModel):
    """
    NovelAI credentials for user account authentication.
    """
    username: str
    password: str

    def __str__(self):
        return f"User(username={self.username})"

    __repr__ = __str__


@dataclass(frozen=True, kw_only=True, slots=True)
class Host:
    """
    Hostnames of NovelAI services and corresponding accepted content types.
    """

    url: str
    accept: str


class AuthError(Exception):
    """
    Exception for account authentication errors.
    """

    pass


class APIError(Exception):
    """
    Exception for package-level errors which need to be fixed in the future development (e.g. validation errors).
    """

    pass


class NovelAIError(Exception):
    """
    Exception for errors returned from NovelAI which are not handled by the package.
    """

    pass


class TimeoutError(NovelAIError):
    """
    Exception for request timeouts.
    """

    pass


class ConcurrentError(NovelAIError):
    """
    Exception for concurrent request errors.
    """

    pass



import os
import asyncio
import unittest
from unittest.mock import AsyncMock, MagicMock

from httpx import AsyncClient

from novelai import (
    NAIClient,
    APIError,
    AuthError,
    NovelAIError,
    ConcurrentError,
    TimeoutError,
    Metadata,
    HEADERS,
    HOSTS,
)


class TestGenerateImage(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.naiclient = NAIClient(
            os.getenv("USERNAME") or "test_username",
            os.getenv("PASSWORD") or "test_password",
        )
        self.naiclient.client = AsyncClient(headers=HEADERS)
        self.metadata = Metadata(prompt="1girl", seed=1, extra_noise_seed=2)

    @unittest.skipIf(
        not (os.getenv("USERNAME") and os.getenv("PASSWORD")),
        "Skipping test_success...",
    )
    async def test_success(self):
        async def task_api():
            with self.subTest("task_api"):
                try:
                    output = await self.naiclient.generate_image(
                        self.metadata, host=HOSTS.API
                    )
                    self.assertTrue("image_0.png" in output)
                except TimeoutError:
                    self.skipTest("task_api timed out")

        async def task_web():
            with self.subTest("task_web"):
                try:
                    output = await self.naiclient.generate_image(
                        self.metadata, host=HOSTS.WEB
                    )
                    self.assertTrue("image_0.png" in output)
                except TimeoutError:
                    self.skipTest("task_web skipped")

        tasks = [
            asyncio.create_task(task_api()),
            asyncio.create_task(task_web()),
        ]
        await asyncio.wait(tasks)

    def test_exceptions(self):
        # Error cases
        error_codes = [400, 401, 402, 409, 429, 500]
        error_exceptions = [
            APIError,
            AuthError,
            AuthError,
            NovelAIError,
            ConcurrentError,
            NovelAIError,
        ]

        for code, exception in zip(error_codes, error_exceptions):
            with self.subTest(f"Status code: {code}"):
                self.naiclient.running = True
                # Mock the AsyncClient's post method to return an error code
                self.naiclient.client.post = AsyncMock(
                    return_value=MagicMock(status_code=code)
                )

                # Check if correct exception is raised
                with self.assertRaises(exception):
                    asyncio.run(self.naiclient.generate_image(self.metadata))


if __name__ == "__main__":
    unittest.main()



import unittest

from novelai import Metadata, RESOLUTIONS


class TestCalculateCost(unittest.TestCase):
    def setUp(self):
        param_s = Metadata(
            prompt="",
            width=RESOLUTIONS.SMALL_PORTRAIT[0],
            height=RESOLUTIONS.SMALL_PORTRAIT[1],
        )
        param_m1 = Metadata(
            prompt="",
            width=RESOLUTIONS.NORMAL_LANDSCAPE[0],
            height=RESOLUTIONS.NORMAL_LANDSCAPE[1],
        )
        param_m2 = Metadata(
            prompt="",
            width=RESOLUTIONS.NORMAL_SQUARE[0],
            height=RESOLUTIONS.NORMAL_SQUARE[1],
            sm=False,
        )
        param_m3 = Metadata(
            prompt="",
            width=RESOLUTIONS.NORMAL_PORTRAIT[0],
            height=RESOLUTIONS.NORMAL_PORTRAIT[1],
            sm_dyn=True,
        )
        param_m4 = Metadata(
            prompt="",
            width=RESOLUTIONS.NORMAL_SQUARE[0],
            height=RESOLUTIONS.NORMAL_SQUARE[1],
            steps=29,
        )
        param_m5 = Metadata(
            prompt="",
            width=RESOLUTIONS.NORMAL_LANDSCAPE[0],
            height=RESOLUTIONS.NORMAL_LANDSCAPE[1],
            n_samples=4,
        )
        param_l1 = Metadata(
            prompt="",
            width=RESOLUTIONS.LARGE_PORTRAIT[0],
            height=RESOLUTIONS.LARGE_PORTRAIT[1],
        )
        param_l2 = Metadata(
            prompt="",
            width=RESOLUTIONS.LARGE_SQUARE[0],
            height=RESOLUTIONS.LARGE_SQUARE[1],
        )
        param_xl = Metadata(
            prompt="",
            width=RESOLUTIONS.WALLPAPER_LANDSCAPE[0],
            height=RESOLUTIONS.WALLPAPER_LANDSCAPE[1],
        )
        self.params = {
            "param_s": param_s,
            "param_m1": param_m1,
            "param_m2": param_m2,
            "param_m3": param_m3,
            "param_m4": param_m4,
            "param_m5": param_m5,
            "param_l1": param_l1,
            "param_l2": param_l2,
            "param_xl": param_xl,
        }
        self.cost_opus = [0, 0, 0, 0, 24, 72, 36, 51, 48]
        self.cost_non_opus = [10, 24, 20, 28, 24, 96, 36, 51, 48]

    def test_opus_true(self):
        for param, cost in zip(self.params, self.cost_opus):
            with self.subTest(f"{param}"):
                result = self.params[param].calculate_cost(is_opus=True)
                self.assertEqual(result, cost)

    def test_opus_false(self):
        for param, cost in zip(self.params, self.cost_non_opus):
            with self.subTest(f"{param}"):
                result = self.params[param].calculate_cost(is_opus=False)
                self.assertEqual(result, cost)


if __name__ == "__main__":
    unittest.main()



from .types import DotDict, Host

HOSTS = DotDict(
    {
        "API": Host(
            url="https://api.novelai.net", accept="application/x-zip-compressed"
        ),
        "WEB": Host(url="https://image.novelai.net", accept="binary/octet-stream"),
    }
)

ENDPOINTS = DotDict(
    {
        "LOGIN": "/user/login",
        "USERDATA": "/user/data",
        "IMAGE": "/ai/generate-image",
    }
)

MODELS = DotDict(
    {
        "V3": "nai-diffusion-3",
        "V3INP": "nai-diffusion-3-inpainting",
    }
)

ACTIONS = DotDict(
    {
        "GENERATE": "generate",
        "INPAINT": "infill",
        "IMG2IMG": "img2img",
    }
)

RESOLUTIONS = DotDict(
    {
        "SMALL_PORTRAIT": (512, 768),
        "SMALL_LANDSCAPE": (768, 512),
        "SMALL_SQUARE": (640, 640),
        "NORMAL_PORTRAIT": (832, 1216),
        "NORMAL_LANDSCAPE": (1216, 832),
        "NORMAL_SQUARE": (1024, 1024),
        "LARGE_PORTRAIT": (1024, 1536),
        "LARGE_LANDSCAPE": (1536, 1024),
        "LARGE_SQUARE": (1472, 1472),
        "WALLPAPER_PORTRAIT": (1088, 1920),
        "WALLPAPER_LANDSCAPE": (1920, 1088),
    }
)

SAMPLERS = DotDict(
    {
        "EULER": "k_euler",
        "EULER_ANC": "k_euler_ancestral",
        "DPM": "k_dpmpp_2s_ancestral",
    }
)

NOISES = DotDict(
    {
        "NATIVE": "native",
        "KARRAS": "karras",
        "EXPONENTIAL": "exponential",
        "POLYEXPONENTIAL": "polyexponential",
    }
)

HEADERS = {
    "Content-Type": "application/json",
    "Origin": "https://novelai.net",
    "Referer": "https://novelai.net",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
}



import io
import zipfile
from hashlib import blake2b
from base64 import urlsafe_b64encode

import argon2

from .types import User


# https://github.com/Aedial/novelai-api/blob/main/novelai_api/utils.py
def encode_access_key(user: User) -> str:
    """
    Generate hashed access key from the user's username and password using the blake2 and argon2 algorithms.

    Parameters
    ----------
    user : `User`
        User object containing username and password

    Returns
    -------
    `str`
        Hashed access key
    """
    pre_salt = f"{user.password[:6]}{user.username}novelai_data_access_key"

    blake = blake2b(digest_size=16)
    blake.update(pre_salt.encode())
    salt = blake.digest()

    raw = argon2.low_level.hash_secret_raw(
        secret=user.password.encode(),
        salt=salt,
        time_cost=2,
        memory_cost=int(2000000 / 1024),
        parallelism=1,
        hash_len=64,
        type=argon2.low_level.Type.ID,
    )
    hashed = urlsafe_b64encode(raw).decode()

    return hashed[:64]


def parse_zip(zip_data: bytes) -> dict:
    """
    Parse binary data of a zip file into a dictionary.

    Parameters
    ----------
    zip_data : `bytes`
        Binary data of a zip file

    Returns
    -------
    `dict`
        Dictionary containing pairs of filename (in `str`) and file data (in `bytes`) for each file in the zip
    """
    files = {}

    with zipfile.ZipFile(io.BytesIO(zip_data)) as zip_file:
        for filename in zip_file.namelist():
            files[filename] = zip_file.read(filename)

    return files
